/*
 * Copyright 2018-2022 Elyra Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import produce from "immer";
import styled from "styled-components";

import { PropertiesPanel, Message } from "./PropertiesPanel";

interface Props {
  selectedNodes?: any[];
  nodes: {
    op: string;
    label?: string;
    app_data: {
      properties?: any;
      parameter_refs?: {
        filehandler?: string;
      };
    };
  }[];
  upstreamNodes?: any[];
  onFileRequested?: (options: any) => any;
  onPropertiesUpdateRequested?: (options: any) => any;
  onChange?: (nodeID: string, data: any) => any;
}

const Heading = styled.div`
  margin-top: 14px;
  padding: 0 47px;
  font-family: ${({ theme }) => theme.typography.fontFamily};
  font-weight: ${({ theme }) => theme.typography.fontWeight};
  font-size: 16px;
  color: ${({ theme }) => theme.palette.text.primary};
  opacity: 0.5;
`;

function NodeProperties({
  selectedNodes,
  nodes,
  upstreamNodes,
  onFileRequested,
  onPropertiesUpdateRequested,
  onChange,
}: Props) {
  if (selectedNodes === undefined || selectedNodes.length === 0) {
    return <Message>Select a node to edit its properties.</Message>;
  }

  if (selectedNodes.length > 1) {
    return (
      <Message>
        Multiple nodes are selected. Select a single node to edit its
        properties.
      </Message>
    );
  }

  const selectedNode = selectedNodes[0];

  if (!selectedNode) {
    return <Message>Select a node to edit its properties.</Message>;
  }

  if (selectedNode.type !== "execution_node") {
    return (
      <Message>This node type doesn't have any editable properties.</Message>
    );
  }

  const nodePropertiesSchema = nodes.find((n) => n.op === selectedNode.op);

  if (nodePropertiesSchema === undefined) {
    return (
      <Message>
        This node uses a component that is not stored in your component
        registry.
        {selectedNode.app_data.component_source !== undefined
          ? ` The component's path is: ${selectedNode.app_data.component_source}`
          : ""}
      </Message>
    );
  }

  if (nodePropertiesSchema?.app_data.properties === undefined) {
    return (
      <Message>This node type doesn't have any editable properties.</Message>
    );
  }

  const refs = nodePropertiesSchema.app_data.parameter_refs;

  // returns the node properties for selectedNode with the most recent content
  const getNodeProperties = (): any => {
    const data: any[] = [];

    // add each upstream node to the data list
    for (const upstreamNode of upstreamNodes ?? []) {
      const nodeDef = nodes.find((n) => n.op === upstreamNode.op);
      const options = [];

      // Add each property with a format of outputpath to the options field
      for (const prop of nodeDef?.app_data.properties.uihints.parameter_info ??
        []) {
        if (prop.data.format === "outputpath") {
          options.push({
            value: prop.parameter_ref,
            label: prop.label.default,
          });
        }
      }
      data.push({
        value: upstreamNode.id,
        label: upstreamNode.app_data?.ui_data?.label,
        options: options,
      });
    }

    // update property data to include data for properties with inputpath format
    return produce(nodePropertiesSchema?.app_data.properties, (draft: any) => {
      for (let prop of draft.uihints.parameter_info) {
        if (prop.data?.format === "inputpath") {
          prop.data = {
            ...prop.data,
            data,
            placeholder: "Select an input source",
          };
        } else if (prop.data?.controls) {
          for (const key in prop.data?.controls) {
            if (prop.data?.controls[key].format === "inputpath") {
              prop.data.controls[key] = {
                ...prop.data.controls[key],
                data,
                placeholder: "Select an input source",
              };
            }
          }
        }
      }
    });
  };

  return (
    <div>
      <Heading>{nodePropertiesSchema.label}</Heading>
      <PropertiesPanel
        refs={refs}
        currentProperties={selectedNode.app_data}
        onPropertiesUpdateRequested={onPropertiesUpdateRequested}
        propertiesSchema={getNodeProperties()}
        onFileRequested={onFileRequested}
        onChange={(data: any) => {
          onChange?.(selectedNode.id, data);
        }}
        id={selectedNode.id}
      />
    </div>
  );
}

export default NodeProperties;
